@@ -39,6 +39,7 @@ class NameApplier : public ExprVisitor::DelegateNop {
   Result EndBlockExpr(BlockExpr*) override;
   Result OnBrExpr(BrExpr*) override;
   Result OnBrIfExpr(BrIfExpr*) override;
+  Result OnBrOnExnExpr(BrOnExnExpr*) override;
   Result OnBrTableExpr(BrTableExpr*) override;
   Result OnCallExpr(CallExpr*) override;
   Result OnCallIndirectExpr(CallIndirectExpr*) override;
@@ -48,17 +49,19 @@ class NameApplier : public ExprVisitor::DelegateNop {
   Result OnGlobalSetExpr(GlobalSetExpr*) override;
   Result BeginIfExpr(IfExpr*) override;
   Result EndIfExpr(IfExpr*) override;
-  Result BeginIfExceptExpr(IfExceptExpr*) override;
-  Result EndIfExceptExpr(IfExceptExpr*) override;
   Result OnLocalGetExpr(LocalGetExpr*) override;
   Result OnLocalSetExpr(LocalSetExpr*) override;
   Result OnLocalTeeExpr(LocalTeeExpr*) override;
   Result BeginLoopExpr(LoopExpr*) override;
   Result EndLoopExpr(LoopExpr*) override;
-  Result OnMemoryDropExpr(MemoryDropExpr*) override;
+  Result OnDataDropExpr(DataDropExpr*) override;
   Result OnMemoryInitExpr(MemoryInitExpr*) override;
-  Result OnTableDropExpr(TableDropExpr*) override;
+  Result OnElemDropExpr(ElemDropExpr*) override;
   Result OnTableInitExpr(TableInitExpr*) override;
+  Result OnTableGetExpr(TableGetExpr*) override;
+  Result OnTableSetExpr(TableSetExpr*) override;
+  Result OnTableGrowExpr(TableGrowExpr*) override;
+  Result OnTableSizeExpr(TableSizeExpr*) override;
   Result BeginTryExpr(TryExpr*) override;
   Result EndTryExpr(TryExpr*) override;
   Result OnThrowExpr(ThrowExpr*) override;
@@ -73,12 +76,13 @@ class NameApplier : public ExprVisitor::DelegateNop {
   Result UseNameForGlobalVar(Var* var);
   Result UseNameForTableVar(Var* var);
   Result UseNameForMemoryVar(Var* var);
-  Result UseNameForExceptVar(Var* var);
+  Result UseNameForEventVar(Var* var);
   Result UseNameForDataSegmentVar(Var* var);
   Result UseNameForElemSegmentVar(Var* var);
   Result UseNameForParamAndLocalVar(Func* func, Var* var);
   Result VisitFunc(Index func_index, Func* func);
   Result VisitGlobal(Global* global);
+  Result VisitEvent(Event* event);
   Result VisitExport(Index export_index, Export* export_);
   Result VisitElemSegment(Index elem_segment_index, ElemSegment* segment);
   Result VisitDataSegment(Index data_segment_index, DataSegment* segment);
@@ -173,12 +177,12 @@ Result NameApplier::UseNameForMemoryVar(Var* var) {
   return Result::Ok;
 }
 
-Result NameApplier::UseNameForExceptVar(Var* var) {
-  Exception* except = module_->GetExcept(*var);
-  if (!except) {
+Result NameApplier::UseNameForEventVar(Var* var) {
+  Event* event = module_->GetEvent(*var);
+  if (!event) {
     return Result::Error;
   }
-  UseNameForVar(except->name, var);
+  UseNameForVar(event->name, var);
   return Result::Ok;
 }
 
@@ -238,7 +242,7 @@ Result NameApplier::EndLoopExpr(LoopExpr* expr) {
   return Result::Ok;
 }
 
-Result NameApplier::OnMemoryDropExpr(MemoryDropExpr* expr) {
+Result NameApplier::OnDataDropExpr(DataDropExpr* expr) {
   CHECK_RESULT(UseNameForDataSegmentVar(&expr->var));
   return Result::Ok;
 }
@@ -248,7 +252,7 @@ Result NameApplier::OnMemoryInitExpr(MemoryInitExpr* expr)  {
   return Result::Ok;
 }
 
-Result NameApplier::OnTableDropExpr(TableDropExpr* expr)  {
+Result NameApplier::OnElemDropExpr(ElemDropExpr* expr)  {
   CHECK_RESULT(UseNameForElemSegmentVar(&expr->var));
   return Result::Ok;
 }
@@ -258,6 +262,26 @@ Result NameApplier::OnTableInitExpr(TableInitExpr* expr)  {
   return Result::Ok;
 }
 
+Result NameApplier::OnTableGetExpr(TableGetExpr* expr)  {
+  CHECK_RESULT(UseNameForTableVar(&expr->var));
+  return Result::Ok;
+}
+
+Result NameApplier::OnTableSetExpr(TableSetExpr* expr)  {
+  CHECK_RESULT(UseNameForTableVar(&expr->var));
+  return Result::Ok;
+}
+
+Result NameApplier::OnTableGrowExpr(TableGrowExpr* expr)  {
+  CHECK_RESULT(UseNameForTableVar(&expr->var));
+  return Result::Ok;
+}
+
+Result NameApplier::OnTableSizeExpr(TableSizeExpr* expr)  {
+  CHECK_RESULT(UseNameForTableVar(&expr->var));
+  return Result::Ok;
+}
+
 Result NameApplier::OnBrExpr(BrExpr* expr) {
   string_view label = FindLabelByVar(&expr->var);
   UseNameForVar(label, &expr->var);
@@ -270,6 +294,13 @@ Result NameApplier::OnBrIfExpr(BrIfExpr* expr) {
   return Result::Ok;
 }
 
+Result NameApplier::OnBrOnExnExpr(BrOnExnExpr* expr) {
+  string_view label = FindLabelByVar(&expr->label_var);
+  UseNameForVar(label, &expr->label_var);
+  CHECK_RESULT(UseNameForEventVar(&expr->event_var));
+  return Result::Ok;
+}
+
 Result NameApplier::OnBrTableExpr(BrTableExpr* expr) {
   for (Var& target : expr->targets) {
     string_view label = FindLabelByVar(&target);
@@ -292,7 +323,7 @@ Result NameApplier::EndTryExpr(TryExpr*) {
 }
 
 Result NameApplier::OnThrowExpr(ThrowExpr* expr) {
-  CHECK_RESULT(UseNameForExceptVar(&expr->var));
+  CHECK_RESULT(UseNameForEventVar(&expr->var));
   return Result::Ok;
 }
 
@@ -305,6 +336,7 @@ Result NameApplier::OnCallIndirectExpr(CallIndirectExpr* expr) {
   if (expr->decl.has_func_type) {
     CHECK_RESULT(UseNameForFuncTypeVar(&expr->decl.type_var));
   }
+  CHECK_RESULT(UseNameForTableVar(&expr->table));
   return Result::Ok;
 }
 
@@ -317,6 +349,7 @@ Result NameApplier::OnReturnCallIndirectExpr(ReturnCallIndirectExpr* expr) {
   if (expr->decl.has_func_type) {
     CHECK_RESULT(UseNameForFuncTypeVar(&expr->decl.type_var));
   }
+  CHECK_RESULT(UseNameForTableVar(&expr->table));
   return Result::Ok;
 }
 
@@ -340,17 +373,6 @@ Result NameApplier::EndIfExpr(IfExpr* expr) {
   return Result::Ok;
 }
 
-Result NameApplier::BeginIfExceptExpr(IfExceptExpr* expr) {
-  PushLabel(expr->true_.label);
-  CHECK_RESULT(UseNameForExceptVar(&expr->except_var));
-  return Result::Ok;
-}
-
-Result NameApplier::EndIfExceptExpr(IfExceptExpr* expr) {
-  PopLabel();
-  return Result::Ok;
-}
-
 Result NameApplier::OnGlobalSetExpr(GlobalSetExpr* expr) {
   CHECK_RESULT(UseNameForGlobalVar(&expr->var));
   return Result::Ok;
@@ -385,6 +407,13 @@ Result NameApplier::VisitGlobal(Global* global) {
   return Result::Ok;
 }
 
+Result NameApplier::VisitEvent(Event* event) {
+  if (event->decl.has_func_type) {
+    CHECK_RESULT(UseNameForFuncTypeVar(&event->decl.type_var));
+  }
+  return Result::Ok;
+}
+
 Result NameApplier::VisitExport(Index export_index, Export* export_) {
   if (export_->kind == ExternalKind::Func) {
     UseNameForFuncVar(&export_->var);
@@ -396,8 +425,10 @@ Result NameApplier::VisitElemSegment(Index elem_segment_index,
                                      ElemSegment* segment) {
   CHECK_RESULT(UseNameForTableVar(&segment->table_var));
   CHECK_RESULT(visitor_.VisitExprList(segment->offset));
-  for (Var& var : segment->vars) {
-    CHECK_RESULT(UseNameForFuncVar(&var));
+  for (ElemExpr& elem_expr : segment->elem_exprs) {
+    if (elem_expr.kind == ElemExprKind::RefFunc) {
+      CHECK_RESULT(UseNameForFuncVar(&elem_expr.var));
+    }
   }
   return Result::Ok;
 }
@@ -415,6 +446,8 @@ Result NameApplier::VisitModule(Module* module) {
     CHECK_RESULT(VisitFunc(i, module->funcs[i]));
   for (size_t i = 0; i < module->globals.size(); ++i)
     CHECK_RESULT(VisitGlobal(module->globals[i]));
+  for (size_t i = 0; i < module->events.size(); ++i)
+    CHECK_RESULT(VisitEvent(module->events[i]));
   for (size_t i = 0; i < module->exports.size(); ++i)
     CHECK_RESULT(VisitExport(i, module->exports[i]));
   for (size_t i = 0; i < module->elem_segments.size(); ++i)
