@@ -115,11 +115,11 @@ class BinaryReaderDelegate {
                                 Index global_index,
                                 Type type,
                                 bool mutable_) = 0;
-  virtual Result OnImportException(Index import_index,
-                                   string_view module_name,
-                                   string_view field_name,
-                                   Index except_index,
-                                   TypeVector& sig) = 0;
+  virtual Result OnImportEvent(Index import_index,
+                               string_view module_name,
+                               string_view field_name,
+                               Index event_index,
+                               Index sig_index) = 0;
   virtual Result EndImportSection() = 0;
 
   /* Function section */
@@ -178,6 +178,7 @@ class BinaryReaderDelegate {
   virtual Result OnOpcodeBare() = 0;
   virtual Result OnOpcodeUint32(uint32_t value) = 0;
   virtual Result OnOpcodeIndex(Index value) = 0;
+  virtual Result OnOpcodeIndexIndex(Index value, Index value2) = 0;
   virtual Result OnOpcodeUint32Uint32(uint32_t value, uint32_t value2) = 0;
   virtual Result OnOpcodeUint64(uint64_t value) = 0;
   virtual Result OnOpcodeF32(uint32_t value) = 0;
@@ -206,11 +207,12 @@ class BinaryReaderDelegate {
   virtual Result OnBlockExpr(Type sig_type) = 0;
   virtual Result OnBrExpr(Index depth) = 0;
   virtual Result OnBrIfExpr(Index depth) = 0;
+  virtual Result OnBrOnExnExpr(Index depth, Index event_index) = 0;
   virtual Result OnBrTableExpr(Index num_targets,
                                Index* target_depths,
                                Index default_target_depth) = 0;
   virtual Result OnCallExpr(Index func_index) = 0;
-  virtual Result OnCallIndirectExpr(Index sig_index) = 0;
+  virtual Result OnCallIndirectExpr(Index sig_index, Index table_index) = 0;
   virtual Result OnCatchExpr() = 0;
   virtual Result OnCompareExpr(Opcode opcode) = 0;
   virtual Result OnConvertExpr(Opcode opcode) = 0;
@@ -226,7 +228,6 @@ class BinaryReaderDelegate {
   virtual Result OnI32ConstExpr(uint32_t value) = 0;
   virtual Result OnI64ConstExpr(uint64_t value) = 0;
   virtual Result OnIfExpr(Type sig_type) = 0;
-  virtual Result OnIfExceptExpr(Type sig_type, Index except_index) = 0;
   virtual Result OnLoadExpr(Opcode opcode,
                             uint32_t alignment_log2,
                             Address offset) = 0;
@@ -235,24 +236,31 @@ class BinaryReaderDelegate {
   virtual Result OnLocalTeeExpr(Index local_index) = 0;
   virtual Result OnLoopExpr(Type sig_type) = 0;
   virtual Result OnMemoryCopyExpr() = 0;
-  virtual Result OnMemoryDropExpr(Index segment_index) = 0;
+  virtual Result OnDataDropExpr(Index segment_index) = 0;
   virtual Result OnMemoryFillExpr() = 0;
   virtual Result OnMemoryGrowExpr() = 0;
   virtual Result OnMemoryInitExpr(Index segment_index) = 0;
   virtual Result OnMemorySizeExpr() = 0;
   virtual Result OnTableCopyExpr() = 0;
-  virtual Result OnTableDropExpr(Index segment_index) = 0;
+  virtual Result OnElemDropExpr(Index segment_index) = 0;
   virtual Result OnTableInitExpr(Index segment_index) = 0;
+  virtual Result OnTableGetExpr(Index table_index) = 0;
+  virtual Result OnTableSetExpr(Index table_index) = 0;
+  virtual Result OnTableGrowExpr(Index table_index) = 0;
+  virtual Result OnTableSizeExpr(Index table_index) = 0;
+  virtual Result OnRefNullExpr() = 0;
+  virtual Result OnRefIsNullExpr() = 0;
   virtual Result OnNopExpr() = 0;
   virtual Result OnRethrowExpr() = 0;
   virtual Result OnReturnExpr() = 0;
   virtual Result OnReturnCallExpr(Index func_index) = 0;
-  virtual Result OnReturnCallIndirectExpr(Index sig_index) = 0;
+  virtual Result OnReturnCallIndirectExpr(Index sig_index,
+                                          Index table_index) = 0;
   virtual Result OnSelectExpr() = 0;
   virtual Result OnStoreExpr(Opcode opcode,
                              uint32_t alignment_log2,
                              Address offset) = 0;
-  virtual Result OnThrowExpr(Index except_index) = 0;
+  virtual Result OnThrowExpr(Index event_index) = 0;
   virtual Result OnTryExpr(Type sig_type) = 0;
 
   virtual Result OnUnaryExpr(Opcode opcode) = 0;
@@ -270,12 +278,14 @@ class BinaryReaderDelegate {
   virtual Result OnElemSegmentCount(Index count) = 0;
   virtual Result BeginElemSegment(Index index,
                                   Index table_index,
-                                  bool passive) = 0;
+                                  bool passive,
+                                  Type elem_type) = 0;
   virtual Result BeginElemSegmentInitExpr(Index index) = 0;
   virtual Result EndElemSegmentInitExpr(Index index) = 0;
-  virtual Result OnElemSegmentFunctionIndexCount(Index index, Index count) = 0;
-  virtual Result OnElemSegmentFunctionIndex(Index segment_index,
-                                            Index func_index) = 0;
+  virtual Result OnElemSegmentElemExprCount(Index index, Index count) = 0;
+  virtual Result OnElemSegmentElemExpr_RefNull(Index segment_index) = 0;
+  virtual Result OnElemSegmentElemExpr_RefFunc(Index segment_index,
+                                               Index func_index) = 0;
   virtual Result EndElemSegment(Index index) = 0;
   virtual Result EndElemSection() = 0;
 
@@ -293,6 +303,11 @@ class BinaryReaderDelegate {
   virtual Result EndDataSegment(Index index) = 0;
   virtual Result EndDataSection() = 0;
 
+  /* DataCount section */
+  virtual Result BeginDataCountSection(Offset size) = 0;
+  virtual Result OnDataCount(Index count) = 0;
+  virtual Result EndDataCountSection() = 0;
+
   /* Names section */
   virtual Result BeginNamesSection(Offset size) = 0;
   virtual Result OnModuleNameSubsection(Index index,
@@ -329,9 +344,9 @@ class BinaryReaderDelegate {
   /* Dylink section */
   virtual Result BeginDylinkSection(Offset size) = 0;
   virtual Result OnDylinkInfo(uint32_t mem_size,
-                              uint32_t mem_align,
+                              uint32_t mem_align_log2,
                               uint32_t table_size,
-                              uint32_t table_align) = 0;
+                              uint32_t table_align_log2) = 0;
   virtual Result OnDylinkNeededCount(Index count) = 0;
   virtual Result OnDylinkNeeded(string_view so_name) = 0;
   virtual Result EndDylinkSection() = 0;
@@ -357,20 +372,24 @@ class BinaryReaderDelegate {
   virtual Result OnSectionSymbol(Index index,
                                  uint32_t flags,
                                  Index section_index) = 0;
+  virtual Result OnEventSymbol(Index index,
+                               uint32_t flags,
+                               string_view name,
+                               Index event_index) = 0;
   virtual Result OnSegmentInfoCount(Index count) = 0;
   virtual Result OnSegmentInfo(Index index,
                                string_view name,
-                               uint32_t alignment,
+                               uint32_t alignment_log2,
                                uint32_t flags) = 0;
   virtual Result OnInitFunctionCount(Index count) = 0;
   virtual Result OnInitFunction(uint32_t priority, Index function_index) = 0;
   virtual Result EndLinkingSection() = 0;
 
-  /* Exception section */
-  virtual Result BeginExceptionSection(Offset size) = 0;
-  virtual Result OnExceptionCount(Index count) = 0;
-  virtual Result OnExceptionType(Index index, TypeVector& sig) = 0;
-  virtual Result EndExceptionSection() = 0;
+  /* Event section */
+  virtual Result BeginEventSection(Offset size) = 0;
+  virtual Result OnEventCount(Index count) = 0;
+  virtual Result OnEventType(Index index, Index sig_index) = 0;
+  virtual Result EndEventSection() = 0;
 
   /* InitExpr - used by elem, data and global sections; these functions are
    * only called between calls to Begin*InitExpr and End*InitExpr */
