@@ -574,48 +574,52 @@ void Environment::Disassemble(Stream* stream,
       }
 
       case Opcode::V128Const: {
-        stream->Writef("%s 0x%08x 0x%08x 0x%08x 0x%08x\n", opcode.GetName(),
+        stream->Writef("%s i32x4 0x%08x 0x%08x 0x%08x 0x%08x\n", opcode.GetName(),
                        ReadU32(&pc), ReadU32(&pc), ReadU32(&pc), ReadU32(&pc));
 
         break;
       }
 
-      case Opcode::MemoryInit:
+      case Opcode::TableGet:
+      case Opcode::TableSet:
+      case Opcode::TableGrow:
+      case Opcode::TableSize:
+      case Opcode::RefNull:
+      case Opcode::RefIsNull:
+      case Opcode::RefFunc:
         WABT_UNREACHABLE;
         break;
 
-      case Opcode::MemoryDrop:
-        WABT_UNREACHABLE;
+      case Opcode::MemoryInit:
+      case Opcode::TableInit: {
+        Index index = ReadU32(&pc);
+        Index segment_index = ReadU32(&pc);
+        stream->Writef("%s $%" PRIindex ", $%" PRIindex
+                       ", %%[-3], %%[-2], %%[-1]\n",
+                       opcode.GetName(), index, segment_index);
+        break;
+      }
+
+      case Opcode::DataDrop:
+      case Opcode::ElemDrop:
+        stream->Writef("%s $%u\n", opcode.GetName(), ReadU32(&pc));
         break;
 
       case Opcode::MemoryCopy:
-        WABT_UNREACHABLE;
-        break;
-
-      case Opcode::MemoryFill:
-        WABT_UNREACHABLE;
-        break;
-
-      case Opcode::TableInit:
-        WABT_UNREACHABLE;
-        break;
-
-      case Opcode::TableDrop:
-        WABT_UNREACHABLE;
-        break;
-
       case Opcode::TableCopy:
-        WABT_UNREACHABLE;
+      case Opcode::MemoryFill:
+        stream->Writef("%s $%u, %%[-3], %%[-2], %%[-1]\n", opcode.GetName(),
+                       ReadU32(&pc));
         break;
 
       // The following opcodes are either never generated or should never be
       // executed.
       case Opcode::Block:
+      case Opcode::BrOnExn:
       case Opcode::Catch:
       case Opcode::Else:
       case Opcode::End:
       case Opcode::If:
-      case Opcode::IfExcept:
       case Opcode::Invalid:
       case Opcode::Loop:
       case Opcode::Rethrow:
