@@ -65,6 +65,12 @@ namespace interp {
   V(TrapHostResultTypeMismatch, "host result type mismatch")                \
   /* we called an import function, but it didn't complete succesfully */    \
   V(TrapHostTrapped, "host function trapped")                               \
+  /* the data segment has been dropped. */                                  \
+  V(TrapDataSegmentDropped, "data segment dropped")                         \
+  /* the element segment has been dropped. */                               \
+  V(TrapElemSegmentDropped, "element segment dropped")                      \
+  /* table access is out of bounds */                                       \
+  V(TrapTableAccessOutOfBounds, "out of bounds table access")               \
   /* we attempted to call a function with the an argument list that doesn't \
    * match the function signature */                                        \
   V(ArgumentTypeMismatch, "argument type mismatch")                         \
@@ -95,9 +101,12 @@ struct FuncSignature {
 };
 
 struct Table {
-  explicit Table(const Limits& limits)
-      : limits(limits), func_indexes(limits.initial, kInvalidIndex) {}
+  explicit Table(Type elem_type, const Limits& limits)
+      : elem_type(elem_type),
+        limits(limits),
+        func_indexes(limits.initial, kInvalidIndex) {}
 
+  Type elem_type;
   Limits limits;
   std::vector<Index> func_indexes;
 };
@@ -111,6 +120,20 @@ struct Memory {
   std::vector<char> data;
 };
 
+struct DataSegment {
+  DataSegment() = default;
+
+  std::vector<char> data;
+  bool dropped = false;
+};
+
+struct ElemSegment {
+  ElemSegment() = default;
+
+  std::vector<Index> elems;
+  bool dropped = false;
+};
+
 // ValueTypeRep converts from one type to its representation on the
 // stack. For example, float -> uint32_t. See Value below.
 template <typename T>
@@ -220,10 +243,10 @@ struct GlobalImport : Import {
   bool mutable_ = false;
 };
 
-struct ExceptImport : Import {
-  ExceptImport() : Import(ExternalKind::Except) {}
-  ExceptImport(string_view module_name, string_view field_name)
-      : Import(ExternalKind::Except, module_name, field_name) {}
+struct EventImport : Import {
+  EventImport() : Import(ExternalKind::Event) {}
+  EventImport(string_view module_name, string_view field_name)
+      : Import(ExternalKind::Event, module_name, field_name) {}
 };
 
 struct Func;
@@ -323,7 +346,7 @@ struct DefinedModule : Module {
   std::vector<TableImport> table_imports;
   std::vector<MemoryImport> memory_imports;
   std::vector<GlobalImport> global_imports;
-  std::vector<ExceptImport> except_imports;
+  std::vector<EventImport> event_imports;
   Index start_func_index; /* kInvalidIndex if not defined */
   IstreamOffset istream_start;
   IstreamOffset istream_end;
@@ -341,7 +364,9 @@ struct HostModule : Module {
   std::pair<HostFunc*, Index> AppendFuncExport(string_view name,
                                                Index sig_index,
                                                HostFunc::Callback);
-  std::pair<Table*, Index> AppendTableExport(string_view name, const Limits&);
+  std::pair<Table*, Index> AppendTableExport(string_view name,
+                                             Type elem_type,
+                                             const Limits&);
   std::pair<Memory*, Index> AppendMemoryExport(string_view name, const Limits&);
   std::pair<Global*, Index> AppendGlobalExport(string_view name,
                                                Type,
@@ -381,6 +406,8 @@ class Environment {
     size_t memories_size = 0;
     size_t tables_size = 0;
     size_t globals_size = 0;
+    size_t data_segments_size = 0;
+    size_t elem_segments_size = 0;
     size_t istream_size = 0;
   };
 
@@ -397,6 +424,8 @@ class Environment {
   Index GetGlobalCount() const { return globals_.size(); }
   Index GetMemoryCount() const { return memories_.size(); }
   Index GetTableCount() const { return tables_.size(); }
+  Index GetDataSegmentCount() const { return data_segments_.size(); }
+  Index GetElemSegmentCount() const { return elem_segments_.size(); }
   Index GetModuleCount() const { return modules_.size(); }
 
   Index GetLastModuleIndex() const {
@@ -421,6 +450,14 @@ class Environment {
     assert(index < tables_.size());
     return &tables_[index];
   }
+  DataSegment* GetDataSegment(Index index) {
+    assert(index < data_segments_.size());
+    return &data_segments_[index];
+  }
+  ElemSegment* GetElemSegment(Index index) {
+    assert(index < elem_segments_.size());
+    return &elem_segments_[index];
+  }
   Module* GetModule(Index index) {
     assert(index < modules_.size());
     return modules_[index].get();
@@ -462,6 +499,18 @@ class Environment {
     return &memories_.back();
   }
 
+  template <typename... Args>
+  DataSegment* EmplaceBackDataSegment(Args&&... args) {
+    data_segments_.emplace_back(std::forward<Args>(args)...);
+    return &data_segments_.back();
+  }
+
+  template <typename... Args>
+  ElemSegment* EmplaceBackElemSegment(Args&&... args) {
+    elem_segments_.emplace_back(std::forward<Args>(args)...);
+    return &elem_segments_.back();
+  }
+
   template <typename... Args>
   Module* EmplaceBackModule(Args&&... args) {
     modules_.emplace_back(std::forward<Args>(args)...);
@@ -497,6 +546,8 @@ class Environment {
   std::vector<Memory> memories_;
   std::vector<Table> tables_;
   std::vector<Global> globals_;
+  std::vector<DataSegment> data_segments_;
+  std::vector<ElemSegment> elem_segments_;
   std::unique_ptr<OutputBuffer> istream_;
   BindingHash module_bindings_;
   BindingHash registered_module_bindings_;
@@ -542,6 +593,11 @@ class Thread {
   template <typename MemType>
   Result GetAtomicAccessAddress(const uint8_t** pc, void** out_address);
 
+  Table* ReadTable(const uint8_t** pc);
+
+  DataSegment* ReadDataSegment(const uint8_t** pc);
+  ElemSegment* ReadElemSegment(const uint8_t** pc);
+
   Value& Top();
   Value& Pick(Index depth);
 
@@ -585,6 +641,14 @@ class Thread {
   template <typename MemType, typename ResultType = MemType>
   Result AtomicRmwCmpxchg(const uint8_t** pc) WABT_WARN_UNUSED;
 
+  Result MemoryInit(const uint8_t** pc) WABT_WARN_UNUSED;
+  Result DataDrop(const uint8_t** pc) WABT_WARN_UNUSED;
+  Result MemoryCopy(const uint8_t** pc) WABT_WARN_UNUSED;
+  Result MemoryFill(const uint8_t** pc) WABT_WARN_UNUSED;
+  Result TableInit(const uint8_t** pc) WABT_WARN_UNUSED;
+  Result ElemDrop(const uint8_t** pc) WABT_WARN_UNUSED;
+  Result TableCopy(const uint8_t** pc) WABT_WARN_UNUSED;
+
   template <typename R, typename T = R>
   Result Unop(UnopFunc<R, T> func) WABT_WARN_UNUSED;
   template <typename R, typename T = R>
@@ -601,6 +665,9 @@ class Thread {
   template <typename T, typename L, typename R, typename P = R>
   Result SimdBinop(BinopFunc<R, P> func) WABT_WARN_UNUSED;
 
+  template <typename T, typename L, typename R, typename P = R>
+  Result SimdRelBinop(BinopFunc<R, P> func) WABT_WARN_UNUSED;
+
   Environment* env_ = nullptr;
   std::vector<Value> value_stack_;
   std::vector<IstreamOffset> call_stack_;
@@ -650,6 +717,8 @@ bool IsArithmeticNan(uint64_t f64_bits);
 std::string TypedValueToString(const TypedValue&);
 const char* ResultToString(Result);
 
+bool ClampToBounds(uint32_t start, uint32_t* length, uint32_t max);
+
 void WriteTypedValue(Stream* stream, const TypedValue&);
 void WriteTypedValues(Stream* stream, const TypedValues&);
 void WriteResult(Stream* stream, const char* desc, Result);
