@@ -48,6 +48,7 @@ class Validator : public ExprVisitor::Delegate {
   Result EndBlockExpr(BlockExpr*) override;
   Result OnBrExpr(BrExpr*) override;
   Result OnBrIfExpr(BrIfExpr*) override;
+  Result OnBrOnExnExpr(BrOnExnExpr*) override;
   Result OnBrTableExpr(BrTableExpr*) override;
   Result OnCallExpr(CallExpr*) override;
   Result OnCallIndirectExpr(CallIndirectExpr*) override;
@@ -60,9 +61,6 @@ class Validator : public ExprVisitor::Delegate {
   Result BeginIfExpr(IfExpr*) override;
   Result AfterIfTrueExpr(IfExpr*) override;
   Result EndIfExpr(IfExpr*) override;
-  Result BeginIfExceptExpr(IfExceptExpr*) override;
-  Result AfterIfExceptTrueExpr(IfExceptExpr*) override;
-  Result EndIfExceptExpr(IfExceptExpr*) override;
   Result OnLoadExpr(LoadExpr*) override;
   Result OnLocalGetExpr(LocalGetExpr*) override;
   Result OnLocalSetExpr(LocalSetExpr*) override;
@@ -70,14 +68,20 @@ class Validator : public ExprVisitor::Delegate {
   Result BeginLoopExpr(LoopExpr*) override;
   Result EndLoopExpr(LoopExpr*) override;
   Result OnMemoryCopyExpr(MemoryCopyExpr*) override;
-  Result OnMemoryDropExpr(MemoryDropExpr*) override;
+  Result OnDataDropExpr(DataDropExpr*) override;
   Result OnMemoryFillExpr(MemoryFillExpr*) override;
   Result OnMemoryGrowExpr(MemoryGrowExpr*) override;
   Result OnMemoryInitExpr(MemoryInitExpr*) override;
   Result OnMemorySizeExpr(MemorySizeExpr*) override;
   Result OnTableCopyExpr(TableCopyExpr*) override;
-  Result OnTableDropExpr(TableDropExpr*) override;
+  Result OnElemDropExpr(ElemDropExpr*) override;
   Result OnTableInitExpr(TableInitExpr*) override;
+  Result OnTableGetExpr(TableGetExpr*) override;
+  Result OnTableSetExpr(TableSetExpr*) override;
+  Result OnTableGrowExpr(TableGrowExpr*) override;
+  Result OnTableSizeExpr(TableSizeExpr*) override;
+  Result OnRefNullExpr(RefNullExpr*) override;
+  Result OnRefIsNullExpr(RefIsNullExpr*) override;
   Result OnNopExpr(NopExpr*) override;
   Result OnReturnExpr(ReturnExpr*) override;
   Result OnReturnCallExpr(ReturnCallExpr*) override;
@@ -168,7 +172,7 @@ class Validator : public ExprVisitor::Delegate {
                                 const char* desc);
   void CheckExprList(const Location* loc, const ExprList& exprs);
   bool CheckHasMemory(const Location* loc, Opcode opcode);
-  bool CheckHasTable(const Location* loc, Opcode opcode);
+  bool CheckHasTable(const Location* loc, Opcode opcode, Index index = 0);
   void CheckHasSharedMemory(const Location* loc, Opcode opcode);
   void CheckBlockDeclaration(const Location* loc,
                              Opcode opcode,
@@ -203,8 +207,8 @@ class Validator : public ExprVisitor::Delegate {
   void CheckAssertReturnNanCommand(const Action* action);
   void CheckCommand(const Command* command);
 
-  void CheckExcept(const Location* loc, const Exception* Except);
-  Result CheckExceptVar(const Var* var, const Exception** out_except);
+  void CheckEvent(const Location* loc, const Event* Event);
+  Result CheckEventVar(const Var* var, const Event** out_event);
 
   const ValidateOptions& options_;
   Errors* errors_ = nullptr;
@@ -215,7 +219,7 @@ class Validator : public ExprVisitor::Delegate {
   Index current_memory_index_ = 0;
   Index current_global_index_ = 0;
   Index num_imported_globals_ = 0;
-  Index current_except_index_ = 0;
+  Index current_event_index_ = 0;
   TypeChecker typechecker_;
   // Cached for access by OnTypecheckerError.
   const Location* expr_loc_ = nullptr;
@@ -493,10 +497,10 @@ bool Validator::CheckHasMemory(const Location* loc, Opcode opcode) {
   return true;
 }
 
-bool Validator::CheckHasTable(const Location* loc, Opcode opcode) {
-  if (current_module_->tables.size() == 0) {
-    PrintError(loc, "%s requires an imported or defined table.",
-               opcode.GetName());
+bool Validator::CheckHasTable(const Location* loc, Opcode opcode, Index index) {
+  if (current_module_->tables.size() <= index) {
+    PrintError(loc, "%s requires table %d to be an imported or defined table.",
+               opcode.GetName(), index);
     return false;
   }
 
@@ -576,6 +580,16 @@ Result Validator::OnBrIfExpr(BrIfExpr* expr) {
   return Result::Ok;
 }
 
+Result Validator::OnBrOnExnExpr(BrOnExnExpr* expr) {
+  expr_loc_ = &expr->loc;
+  const Event* event;
+  if (Succeeded(CheckEventVar(&expr->event_var, &event))) {
+    typechecker_.OnBrOnExn(expr->label_var.index(),
+                           event->decl.sig.param_types);
+  }
+  return Result::Ok;
+}
+
 Result Validator::OnBrTableExpr(BrTableExpr* expr) {
   expr_loc_ = &expr->loc;
   typechecker_.BeginBrTable();
@@ -599,9 +613,7 @@ Result Validator::OnCallExpr(CallExpr* expr) {
 
 Result Validator::OnCallIndirectExpr(CallIndirectExpr* expr) {
   expr_loc_ = &expr->loc;
-  if (current_module_->tables.size() == 0) {
-    PrintError(&expr->loc, "found call_indirect operator, but no table");
-  }
+  CheckHasTable(&expr->loc, Opcode::CallIndirect, expr->table.index());
   CheckFuncSignature(&expr->loc, expr->decl);
   typechecker_.OnCallIndirect(expr->decl.sig.param_types,
                               expr->decl.sig.result_types);
@@ -677,33 +689,6 @@ Result Validator::EndIfExpr(IfExpr* expr) {
   return Result::Ok;
 }
 
-Result Validator::BeginIfExceptExpr(IfExceptExpr* expr) {
-  expr_loc_ = &expr->loc;
-  CheckBlockDeclaration(&expr->loc, Opcode::IfExcept, &expr->true_.decl);
-  const Exception* except;
-  TypeVector except_sig;
-  if (Succeeded(CheckExceptVar(&expr->except_var, &except))) {
-    except_sig = except->sig;
-  }
-  typechecker_.OnIfExcept(expr->true_.decl.sig.param_types,
-                          expr->true_.decl.sig.result_types, except_sig);
-  return Result::Ok;
-}
-
-Result Validator::AfterIfExceptTrueExpr(IfExceptExpr* expr) {
-  if (!expr->false_.empty()) {
-    typechecker_.OnElse();
-  }
-  return Result::Ok;
-}
-
-Result Validator::EndIfExceptExpr(IfExceptExpr* expr) {
-  expr_loc_ =
-      expr->false_.empty() ? &expr->true_.end_loc : &expr->false_end_loc;
-  typechecker_.OnEnd();
-  return Result::Ok;
-}
-
 Result Validator::OnLoadExpr(LoadExpr* expr) {
   expr_loc_ = &expr->loc;
   CheckHasMemory(&expr->loc, expr->opcode);
@@ -752,11 +737,11 @@ Result Validator::OnMemoryCopyExpr(MemoryCopyExpr* expr) {
   return Result::Ok;
 }
 
-Result Validator::OnMemoryDropExpr(MemoryDropExpr* expr) {
+Result Validator::OnDataDropExpr(DataDropExpr* expr) {
   expr_loc_ = &expr->loc;
-  CheckHasMemory(&expr->loc, Opcode::MemoryDrop);
+  CheckHasMemory(&expr->loc, Opcode::DataDrop);
   CheckDataSegmentVar(&expr->var);
-  typechecker_.OnMemoryDrop(expr->var.index());
+  typechecker_.OnDataDrop(expr->var.index());
   return Result::Ok;
 }
 
@@ -796,11 +781,11 @@ Result Validator::OnTableCopyExpr(TableCopyExpr* expr) {
   return Result::Ok;
 }
 
-Result Validator::OnTableDropExpr(TableDropExpr* expr) {
+Result Validator::OnElemDropExpr(ElemDropExpr* expr) {
   expr_loc_ = &expr->loc;
-  CheckHasTable(&expr->loc, Opcode::TableDrop);
+  CheckHasTable(&expr->loc, Opcode::ElemDrop);
   CheckElemSegmentVar(&expr->var);
-  typechecker_.OnTableDrop(expr->var.index());
+  typechecker_.OnElemDrop(expr->var.index());
   return Result::Ok;
 }
 
@@ -812,6 +797,46 @@ Result Validator::OnTableInitExpr(TableInitExpr* expr) {
   return Result::Ok;
 }
 
+Result Validator::OnTableGetExpr(TableGetExpr* expr) {
+  expr_loc_ = &expr->loc;
+  CheckHasTable(&expr->loc, Opcode::TableGet, expr->var.index());
+  typechecker_.OnTableGet(expr->var.index());
+  return Result::Ok;
+}
+
+Result Validator::OnTableSetExpr(TableSetExpr* expr) {
+  expr_loc_ = &expr->loc;
+  CheckHasTable(&expr->loc, Opcode::TableSet, expr->var.index());
+  typechecker_.OnTableSet(expr->var.index());
+  return Result::Ok;
+}
+
+Result Validator::OnTableGrowExpr(TableGrowExpr* expr) {
+  expr_loc_ = &expr->loc;
+  CheckHasTable(&expr->loc, Opcode::TableGrow, expr->var.index());
+  typechecker_.OnTableGrow(expr->var.index());
+  return Result::Ok;
+}
+
+Result Validator::OnTableSizeExpr(TableSizeExpr* expr) {
+  expr_loc_ = &expr->loc;
+  CheckHasTable(&expr->loc, Opcode::TableSize, expr->var.index());
+  typechecker_.OnTableSize(expr->var.index());
+  return Result::Ok;
+}
+
+Result Validator::OnRefNullExpr(RefNullExpr* expr) {
+  expr_loc_ = &expr->loc;
+  typechecker_.OnRefNullExpr();
+  return Result::Ok;
+}
+
+Result Validator::OnRefIsNullExpr(RefIsNullExpr* expr) {
+  expr_loc_ = &expr->loc;
+  typechecker_.OnRefIsNullExpr();
+  return Result::Ok;
+}
+
 Result Validator::OnNopExpr(NopExpr* expr) {
   expr_loc_ = &expr->loc;
   return Result::Ok;
@@ -835,9 +860,7 @@ Result Validator::OnReturnCallExpr(ReturnCallExpr* expr) {
 
 Result Validator::OnReturnCallIndirectExpr(ReturnCallIndirectExpr* expr) {
   expr_loc_ = &expr->loc;
-  if (current_module_->tables.empty()) {
-    PrintError(&expr->loc, "found return_call_indirect operator, but no table");
-  }
+  CheckHasTable(&expr->loc, Opcode::ReturnCallIndirect, expr->table.index());
   CheckFuncSignature(&expr->loc, expr->decl);
   typechecker_.OnReturnCallIndirect(expr->decl.sig.param_types,
                               expr->decl.sig.result_types);
@@ -892,9 +915,9 @@ Result Validator::EndTryExpr(TryExpr* expr) {
 
 Result Validator::OnThrowExpr(ThrowExpr* expr) {
   expr_loc_ = &expr->loc;
-  const Exception* except;
-  if (Succeeded(CheckExceptVar(&expr->var, &except))) {
-    typechecker_.OnThrow(except->sig);
+  const Event* event;
+  if (Succeeded(CheckEventVar(&expr->var, &event))) {
+    typechecker_.OnThrow(event->decl.sig.param_types);
   }
   return Result::Ok;
 }
@@ -1074,7 +1097,8 @@ void Validator::CheckLimits(const Location* loc,
 }
 
 void Validator::CheckTable(const Location* loc, const Table* table) {
-  if (current_table_index_ == 1) {
+  if (current_table_index_ == 1 &&
+      !options_.features.reference_types_enabled()) {
     PrintError(loc, "only one table allowed");
   }
   CheckLimits(loc, &table->elem_limits, UINT32_MAX, "elems");
@@ -1082,6 +1106,13 @@ void Validator::CheckTable(const Location* loc, const Table* table) {
   if (table->elem_limits.is_shared) {
     PrintError(loc, "tables may not be shared");
   }
+  if (table->elem_type == Type::Anyref &&
+      !options_.features.reference_types_enabled()) {
+    PrintError(loc, "tables must have anyref type");
+  }
+  if (table->elem_type != Type::Anyref && table->elem_type != Type::Funcref) {
+    PrintError(loc, "tables must have anyref or funcref type");
+  }
 }
 
 void Validator::CheckElemSegments(const Module* module) {
@@ -1089,8 +1120,10 @@ void Validator::CheckElemSegments(const Module* module) {
     if (auto elem_segment_field = dyn_cast<ElemSegmentModuleField>(&field)) {
       auto&& elem_segment = elem_segment_field->elem_segment;
       const Table* table;
-      for (const Var& var : elem_segment.vars) {
-        CheckFuncVar(&var, nullptr);
+      for (const ElemExpr& elem_expr : elem_segment.elem_exprs) {
+        if (elem_expr.kind == ElemExprKind::RefFunc) {
+          CheckFuncVar(&elem_expr.var, nullptr);
+        }
       }
 
       if (elem_segment.passive)  {
@@ -1139,9 +1172,9 @@ void Validator::CheckDataSegments(const Module* module) {
 
 void Validator::CheckImport(const Location* loc, const Import* import) {
   switch (import->kind()) {
-    case ExternalKind::Except:
-      ++current_except_index_;
-      CheckExcept(loc, &cast<ExceptionImport>(import)->except);
+    case ExternalKind::Event:
+      ++current_event_index_;
+      CheckEvent(loc, &cast<EventImport>(import)->event);
       break;
 
     case ExternalKind::Func: {
@@ -1177,8 +1210,8 @@ void Validator::CheckImport(const Location* loc, const Import* import) {
 
 void Validator::CheckExport(const Location* loc, const Export* export_) {
   switch (export_->kind) {
-    case ExternalKind::Except:
-      CheckExceptVar(&export_->var, nullptr);
+    case ExternalKind::Event:
+      CheckEventVar(&export_->var, nullptr);
       break;
     case ExternalKind::Func:
       CheckFuncVar(&export_->var, nullptr);
@@ -1221,13 +1254,13 @@ Result Validator::CheckModule(const Module* module) {
   current_memory_index_ = 0;
   current_global_index_ = 0;
   num_imported_globals_ = 0;
-  current_except_index_ = 0;
+  current_event_index_ = 0;
 
   for (const ModuleField& field : module->fields) {
     switch (field.type()) {
-      case ModuleFieldType::Except:
-        ++current_except_index_;
-        CheckExcept(&field.loc, &cast<ExceptionModuleField>(&field)->except);
+      case ModuleFieldType::Event:
+        ++current_event_index_;
+        CheckEvent(&field.loc, &cast<EventModuleField>(&field)->event);
         break;
 
       case ModuleFieldType::Func:
@@ -1363,29 +1396,19 @@ Result Validator::CheckGet(const GetAction* action, Type* out_type) {
   return Result::Ok;
 }
 
-Result Validator::CheckExceptVar(const Var* var, const Exception** out_except) {
+Result Validator::CheckEventVar(const Var* var, const Event** out_event) {
   Index index;
-  CHECK_RESULT(
-      CheckVar(current_module_->excepts.size(), var, "except", &index));
-  if (out_except) {
-    *out_except = current_module_->excepts[index];
+  CHECK_RESULT(CheckVar(current_module_->events.size(), var, "event", &index));
+  if (out_event) {
+    *out_event = current_module_->events[index];
   }
   return Result::Ok;
 }
 
-void Validator::CheckExcept(const Location* loc, const Exception* except) {
-  for (Type ty : except->sig) {
-    switch (ty) {
-      case Type::I32:
-      case Type::I64:
-      case Type::F32:
-      case Type::F64:
-      case Type::V128:
-        break;
-      default:
-        PrintError(loc, "Invalid exception type: %s", GetTypeName(ty));
-        break;
-    }
+void Validator::CheckEvent(const Location* loc, const Event* event) {
+  CheckFuncSignature(loc, event->decl);
+  if (event->decl.sig.GetNumResults() > 0) {
+    PrintError(loc, "Event signature must have 0 results.");
   }
 }
 
@@ -1531,12 +1554,6 @@ class Validator::CheckFuncSignatureExprVisitorDelegate
     return Result::Ok;
   }
 
-  Result BeginIfExceptExpr(IfExceptExpr* expr) override {
-    validator_->CheckBlockDeclaration(&expr->loc, Opcode::IfExcept,
-                                      &expr->true_.decl);
-    return Result::Ok;
-  }
-
   Result BeginLoopExpr(LoopExpr* expr) override {
     validator_->CheckBlockDeclaration(&expr->loc, Opcode::Loop,
                                       &expr->block.decl);
