@@ -17,6 +17,7 @@
 #include "src/wat-writer.h"
 
 #include <algorithm>
+#include <array>
 #include <cassert>
 #include <cinttypes>
 #include <cstdarg>
@@ -141,7 +142,6 @@ class WatWriter {
   void WriteBeginBlock(LabelType label_type,
                        const Block& block,
                        const char* text);
-  void WriteBeginIfExceptBlock(const IfExceptExpr* expr);
   void WriteEndBlock();
   void WriteConst(const Const& const_);
   void WriteExpr(const Expr* expr);
@@ -158,8 +158,7 @@ class WatWriter {
   void WriteFunc(const Func& func);
   void WriteBeginGlobal(const Global& global);
   void WriteGlobal(const Global& global);
-  void WriteBeginException(const Exception& except);
-  void WriteException(const Exception& except);
+  void WriteEvent(const Event& event);
   void WriteLimits(const Limits& limits);
   void WriteTable(const Table& table);
   void WriteElemSegment(const ElemSegment& segment);
@@ -208,7 +207,7 @@ class WatWriter {
   Index table_index_ = 0;
   Index memory_index_ = 0;
   Index func_type_index_ = 0;
-  Index except_index_ = 0;
+  Index event_index_ = 0;
   Index data_segment_index_ = 0;
   Index elem_segment_index_ = 0;
 };
@@ -443,26 +442,6 @@ void WatWriter::WriteBeginBlock(LabelType label_type,
   Indent();
 }
 
-void WatWriter::WriteBeginIfExceptBlock(const IfExceptExpr* expr) {
-  const Block& block = expr->true_;
-  WritePutsSpace(Opcode::IfExcept_Opcode.GetName());
-  bool has_label = !block.label.empty();
-  if (has_label) {
-    WriteString(block.label, NextChar::Space);
-  }
-  WriteTypes(block.decl.sig.param_types, "param");
-  WriteTypes(block.decl.sig.result_types, "result");
-  WriteVar(expr->except_var, NextChar::Space);
-  if (!has_label) {
-    Writef(" ;; label = @%" PRIindex, GetLabelStackSize());
-  }
-  WriteNewline(FORCE_NEWLINE);
-  label_stack_.emplace_back(LabelType::IfExcept, block.label,
-                            block.decl.sig.param_types,
-                            block.decl.sig.result_types);
-  Indent();
-}
-
 void WatWriter::WriteEndBlock() {
   Dedent();
   label_stack_.pop_back();
@@ -509,7 +488,7 @@ void WatWriter::WriteConst(const Const& const_) {
 
     case Type::V128: {
       WritePutsSpace(Opcode::V128Const_Opcode.GetName());
-      Writef("i32 0x%08x 0x%08x 0x%08x 0x%08x", const_.v128_bits.v[0],
+      Writef("i32x4 0x%08x 0x%08x 0x%08x 0x%08x", const_.v128_bits.v[0],
              const_.v128_bits.v[1], const_.v128_bits.v[2],
              const_.v128_bits.v[3]);
       WriteNewline(NO_FORCE_NEWLINE);
@@ -544,6 +523,7 @@ class WatWriter::ExprVisitorDelegate : public ExprVisitor::Delegate {
   Result EndBlockExpr(BlockExpr*) override;
   Result OnBrExpr(BrExpr*) override;
   Result OnBrIfExpr(BrIfExpr*) override;
+  Result OnBrOnExnExpr(BrOnExnExpr*) override;
   Result OnBrTableExpr(BrTableExpr*) override;
   Result OnCallExpr(CallExpr*) override;
   Result OnCallIndirectExpr(CallIndirectExpr*) override;
@@ -556,9 +536,6 @@ class WatWriter::ExprVisitorDelegate : public ExprVisitor::Delegate {
   Result BeginIfExpr(IfExpr*) override;
   Result AfterIfTrueExpr(IfExpr*) override;
   Result EndIfExpr(IfExpr*) override;
-  Result BeginIfExceptExpr(IfExceptExpr*) override;
-  Result AfterIfExceptTrueExpr(IfExceptExpr*) override;
-  Result EndIfExceptExpr(IfExceptExpr*) override;
   Result OnLoadExpr(LoadExpr*) override;
   Result OnLocalGetExpr(LocalGetExpr*) override;
   Result OnLocalSetExpr(LocalSetExpr*) override;
@@ -566,14 +543,20 @@ class WatWriter::ExprVisitorDelegate : public ExprVisitor::Delegate {
   Result BeginLoopExpr(LoopExpr*) override;
   Result EndLoopExpr(LoopExpr*) override;
   Result OnMemoryCopyExpr(MemoryCopyExpr*) override;
-  Result OnMemoryDropExpr(MemoryDropExpr*) override;
+  Result OnDataDropExpr(DataDropExpr*) override;
   Result OnMemoryFillExpr(MemoryFillExpr*) override;
   Result OnMemoryGrowExpr(MemoryGrowExpr*) override;
   Result OnMemoryInitExpr(MemoryInitExpr*) override;
   Result OnMemorySizeExpr(MemorySizeExpr*) override;
   Result OnTableCopyExpr(TableCopyExpr*) override;
-  Result OnTableDropExpr(TableDropExpr*) override;
+  Result OnElemDropExpr(ElemDropExpr*) override;
   Result OnTableInitExpr(TableInitExpr*) override;
+  Result OnTableGetExpr(TableGetExpr*) override;
+  Result OnTableSetExpr(TableSetExpr*) override;
+  Result OnTableGrowExpr(TableGrowExpr*) override;
+  Result OnTableSizeExpr(TableSizeExpr*) override;
+  Result OnRefNullExpr(RefNullExpr*) override;
+  Result OnRefIsNullExpr(RefIsNullExpr*) override;
   Result OnNopExpr(NopExpr*) override;
   Result OnReturnExpr(ReturnExpr*) override;
   Result OnReturnCallExpr(ReturnCallExpr*) override;
@@ -629,6 +612,13 @@ Result WatWriter::ExprVisitorDelegate::OnBrIfExpr(BrIfExpr* expr) {
   return Result::Ok;
 }
 
+Result WatWriter::ExprVisitorDelegate::OnBrOnExnExpr(BrOnExnExpr* expr) {
+  writer_->WritePutsSpace(Opcode::BrOnExn_Opcode.GetName());
+  writer_->WriteBrVar(expr->label_var, NextChar::Space);
+  writer_->WriteVar(expr->event_var, NextChar::Newline);
+  return Result::Ok;
+}
+
 Result WatWriter::ExprVisitorDelegate::OnBrTableExpr(BrTableExpr* expr) {
   writer_->WritePutsSpace(Opcode::BrTable_Opcode.GetName());
   for (const Var& var : expr->targets) {
@@ -649,7 +639,13 @@ Result WatWriter::ExprVisitorDelegate::OnCallIndirectExpr(
   writer_->WritePutsSpace(Opcode::CallIndirect_Opcode.GetName());
   writer_->WriteOpenSpace("type");
   writer_->WriteVar(expr->decl.type_var, NextChar::Space);
-  writer_->WriteCloseNewline();
+
+  if (expr->table.is_index() && expr->table.index() == 0) {
+    writer_->WriteCloseNewline();
+  } else {
+    writer_->WriteCloseSpace();
+    writer_->WriteVar(expr->table, NextChar::Newline);
+  }
   return Result::Ok;
 }
 
@@ -706,29 +702,6 @@ Result WatWriter::ExprVisitorDelegate::EndIfExpr(IfExpr* expr) {
   return Result::Ok;
 }
 
-Result WatWriter::ExprVisitorDelegate::BeginIfExceptExpr(IfExceptExpr* expr) {
-  // Can't use WriteBeginBlock because if_except has an additional exception
-  // index argument.
-  writer_->WriteBeginIfExceptBlock(expr);
-  return Result::Ok;
-}
-
-Result WatWriter::ExprVisitorDelegate::AfterIfExceptTrueExpr(
-    IfExceptExpr* expr) {
-  if (!expr->false_.empty()) {
-    writer_->Dedent();
-    writer_->WritePutsSpace(Opcode::Else_Opcode.GetName());
-    writer_->Indent();
-    writer_->WriteNewline(FORCE_NEWLINE);
-  }
-  return Result::Ok;
-}
-
-Result WatWriter::ExprVisitorDelegate::EndIfExceptExpr(IfExceptExpr* expr) {
-  writer_->WriteEndBlock();
-  return Result::Ok;
-}
-
 Result WatWriter::ExprVisitorDelegate::OnLoadExpr(LoadExpr* expr) {
   writer_->WriteLoadStoreExpr<LoadExpr>(expr);
   return Result::Ok;
@@ -768,8 +741,8 @@ Result WatWriter::ExprVisitorDelegate::OnMemoryCopyExpr(MemoryCopyExpr* expr) {
   return Result::Ok;
 }
 
-Result WatWriter::ExprVisitorDelegate::OnMemoryDropExpr(MemoryDropExpr* expr) {
-  writer_->WritePutsSpace(Opcode::MemoryDrop_Opcode.GetName());
+Result WatWriter::ExprVisitorDelegate::OnDataDropExpr(DataDropExpr* expr) {
+  writer_->WritePutsSpace(Opcode::DataDrop_Opcode.GetName());
   writer_->WriteVar(expr->var, NextChar::Newline);
   return Result::Ok;
 }
@@ -784,14 +757,14 @@ Result WatWriter::ExprVisitorDelegate::OnMemoryGrowExpr(MemoryGrowExpr* expr) {
   return Result::Ok;
 }
 
-Result WatWriter::ExprVisitorDelegate::OnMemoryInitExpr(MemoryInitExpr* expr) {
-  writer_->WritePutsSpace(Opcode::MemoryInit_Opcode.GetName());
-  writer_->WriteVar(expr->var, NextChar::Newline);
+Result WatWriter::ExprVisitorDelegate::OnMemorySizeExpr(MemorySizeExpr* expr) {
+  writer_->WritePutsNewline(Opcode::MemorySize_Opcode.GetName());
   return Result::Ok;
 }
 
-Result WatWriter::ExprVisitorDelegate::OnMemorySizeExpr(MemorySizeExpr* expr) {
-  writer_->WritePutsNewline(Opcode::MemorySize_Opcode.GetName());
+Result WatWriter::ExprVisitorDelegate::OnMemoryInitExpr(MemoryInitExpr* expr) {
+  writer_->WritePutsSpace(Opcode::MemoryInit_Opcode.GetName());
+  writer_->WriteVar(expr->var, NextChar::Newline);
   return Result::Ok;
 }
 
@@ -800,8 +773,8 @@ Result WatWriter::ExprVisitorDelegate::OnTableCopyExpr(TableCopyExpr* expr) {
   return Result::Ok;
 }
 
-Result WatWriter::ExprVisitorDelegate::OnTableDropExpr(TableDropExpr* expr) {
-  writer_->WritePutsSpace(Opcode::TableDrop_Opcode.GetName());
+Result WatWriter::ExprVisitorDelegate::OnElemDropExpr(ElemDropExpr* expr) {
+  writer_->WritePutsSpace(Opcode::ElemDrop_Opcode.GetName());
   writer_->WriteVar(expr->var, NextChar::Newline);
   return Result::Ok;
 }
@@ -812,6 +785,40 @@ Result WatWriter::ExprVisitorDelegate::OnTableInitExpr(TableInitExpr* expr) {
   return Result::Ok;
 }
 
+Result WatWriter::ExprVisitorDelegate::OnTableGetExpr(TableGetExpr* expr) {
+  writer_->WritePutsSpace(Opcode::TableGet_Opcode.GetName());
+  writer_->WriteVar(expr->var, NextChar::Newline);
+  return Result::Ok;
+}
+
+Result WatWriter::ExprVisitorDelegate::OnTableSetExpr(TableSetExpr* expr) {
+  writer_->WritePutsSpace(Opcode::TableSet_Opcode.GetName());
+  writer_->WriteVar(expr->var, NextChar::Newline);
+  return Result::Ok;
+}
+
+Result WatWriter::ExprVisitorDelegate::OnTableGrowExpr(TableGrowExpr* expr) {
+  writer_->WritePutsSpace(Opcode::TableGrow_Opcode.GetName());
+  writer_->WriteVar(expr->var, NextChar::Newline);
+  return Result::Ok;
+}
+
+Result WatWriter::ExprVisitorDelegate::OnTableSizeExpr(TableSizeExpr* expr) {
+  writer_->WritePutsSpace(Opcode::TableSize_Opcode.GetName());
+  writer_->WriteVar(expr->var, NextChar::Newline);
+  return Result::Ok;
+}
+
+Result WatWriter::ExprVisitorDelegate::OnRefNullExpr(RefNullExpr* expr) {
+  writer_->WritePutsNewline(Opcode::RefNull_Opcode.GetName());
+  return Result::Ok;
+}
+
+Result WatWriter::ExprVisitorDelegate::OnRefIsNullExpr(RefIsNullExpr* expr) {
+  writer_->WritePutsNewline(Opcode::RefIsNull_Opcode.GetName());
+  return Result::Ok;
+}
+
 Result WatWriter::ExprVisitorDelegate::OnNopExpr(NopExpr* expr) {
   writer_->WritePutsNewline(Opcode::Nop_Opcode.GetName());
   return Result::Ok;
@@ -937,8 +944,10 @@ Result WatWriter::ExprVisitorDelegate::OnSimdLaneOpExpr(SimdLaneOpExpr* expr) {
 Result WatWriter::ExprVisitorDelegate::OnSimdShuffleOpExpr(
     SimdShuffleOpExpr* expr) {
   writer_->WritePutsSpace(expr->opcode.GetName());
-  writer_->Writef(" 0x%08x 0x%08x 0x%08x 0x%08x", (expr->val.v[0]), (expr->val.v[1]),
-                  (expr->val.v[2]), (expr->val.v[3]));
+  std::array<uint8_t, 16> values = Bitcast<std::array<uint8_t, 16>>(expr->val);
+  for (int32_t lane = 0; lane < 16; ++lane) {
+    writer_->Writef(" %u", values[lane]);
+  }
   writer_->WritePutsNewline("");
   return Result::Ok;
 }
@@ -999,11 +1008,13 @@ void WatWriter::WriteFoldedExpr(const Expr* expr) {
     case ExprType::AtomicRmw:
     case ExprType::Binary:
     case ExprType::Compare:
+    case ExprType::TableGrow:
       PushExpr(expr, 2, 1);
       break;
 
     case ExprType::AtomicStore:
     case ExprType::Store:
+    case ExprType::TableSet:
       PushExpr(expr, 2, 0);
       break;
 
@@ -1021,6 +1032,10 @@ void WatWriter::WriteFoldedExpr(const Expr* expr) {
       break;
     }
 
+    case ExprType::BrOnExn:
+      PushExpr(expr, 1, 1);
+      break;
+
     case ExprType::BrTable:
       PushExpr(expr, GetLabelArity(cast<BrTableExpr>(expr)->default_target) + 1,
                1);
@@ -1056,12 +1071,14 @@ void WatWriter::WriteFoldedExpr(const Expr* expr) {
     case ExprType::GlobalGet:
     case ExprType::LocalGet:
     case ExprType::MemorySize:
+    case ExprType::TableSize:
     case ExprType::Unreachable:
+    case ExprType::RefNull:
       PushExpr(expr, 0, 1);
       break;
 
-    case ExprType::MemoryDrop:
-    case ExprType::TableDrop:
+    case ExprType::DataDrop:
+    case ExprType::ElemDrop:
       PushExpr(expr, 0, 0);
       break;
 
@@ -1079,6 +1096,8 @@ void WatWriter::WriteFoldedExpr(const Expr* expr) {
     case ExprType::LocalTee:
     case ExprType::MemoryGrow:
     case ExprType::Unary:
+    case ExprType::TableGet:
+    case ExprType::RefIsNull:
       PushExpr(expr, 1, 1);
       break;
 
@@ -1092,11 +1111,6 @@ void WatWriter::WriteFoldedExpr(const Expr* expr) {
       PushExpr(expr, 1, cast<IfExpr>(expr)->true_.decl.sig.GetNumResults());
       break;
 
-    case ExprType::IfExcept:
-      PushExpr(expr, 1,
-               cast<IfExceptExpr>(expr)->true_.decl.sig.GetNumResults());
-      break;
-
     case ExprType::Loop:
       PushExpr(expr, 0, cast<LoopExpr>(expr)->block.decl.sig.GetNumResults());
       break;
@@ -1122,8 +1136,8 @@ void WatWriter::WriteFoldedExpr(const Expr* expr) {
     case ExprType::Throw: {
       auto throw_ = cast<ThrowExpr>(expr);
       Index operand_count = 0;
-      if (Exception* except = module_->GetExcept(throw_->var)) {
-        operand_count = except->sig.size();
+      if (Event* event = module_->GetEvent(throw_->var)) {
+        operand_count = event->decl.sig.param_types.size();
       }
       PushExpr(expr, operand_count, 0);
       break;
@@ -1273,25 +1287,6 @@ void WatWriter::FlushExprTree(const ExprTree& expr_tree) {
       break;
     }
 
-    case ExprType::IfExcept: {
-      auto if_except_expr = cast<IfExceptExpr>(expr_tree.expr);
-      WritePuts("(", NextChar::None);
-      WriteBeginIfExceptBlock(if_except_expr);
-      FlushExprTreeVector(expr_tree.children);
-      WriteOpenNewline("then");
-      WriteFoldedExprList(if_except_expr->true_.exprs);
-      FlushExprTreeStack();
-      WriteCloseNewline();
-      if (!if_except_expr->false_.empty()) {
-        WriteOpenNewline("else");
-        WriteFoldedExprList(if_except_expr->false_);
-        FlushExprTreeStack();
-        WriteCloseNewline();
-      }
-      WriteCloseNewline();
-      break;
-    }
-
     case ExprType::Try: {
       auto try_expr = cast<TryExpr>(expr_tree.expr);
       WritePuts("(", NextChar::None);
@@ -1453,17 +1448,18 @@ void WatWriter::WriteGlobal(const Global& global) {
   WriteCloseNewline();
 }
 
-void WatWriter::WriteBeginException(const Exception& except) {
-  WriteOpenSpace("except");
-  WriteNameOrIndex(except.name, except_index_, NextChar::Space);
-  WriteInlineExports(ExternalKind::Except, except_index_);
-  WriteInlineImport(ExternalKind::Except, except_index_);
-  WriteTypes(except.sig, nullptr);
-  ++except_index_;
-}
-
-void WatWriter::WriteException(const Exception& except) {
-  WriteBeginException(except);
+void WatWriter::WriteEvent(const Event& event) {
+  WriteOpenSpace("event");
+  WriteNameOrIndex(event.name, event_index_, NextChar::Space);
+  WriteInlineExports(ExternalKind::Event, event_index_);
+  WriteInlineImport(ExternalKind::Event, event_index_);
+  if (event.decl.has_func_type) {
+    WriteOpenSpace("type");
+    WriteVar(event.decl.type_var, NextChar::None);
+    WriteCloseSpace();
+  }
+  WriteTypes(event.decl.sig.param_types, "param");
+  ++event_index_;
   WriteCloseNewline();
 }
 
@@ -1483,7 +1479,7 @@ void WatWriter::WriteTable(const Table& table) {
   WriteInlineExports(ExternalKind::Table, table_index_);
   WriteInlineImport(ExternalKind::Table, table_index_);
   WriteLimits(table.elem_limits);
-  WritePutsSpace("anyfunc");
+  WriteType(table.elem_type, NextChar::None);
   WriteCloseNewline();
   table_index_++;
 }
@@ -1493,11 +1489,24 @@ void WatWriter::WriteElemSegment(const ElemSegment& segment) {
   WriteNameOrIndex(segment.name, elem_segment_index_, NextChar::Space);
   if (segment.passive) {
     WritePutsSpace("passive");
+    WriteType(segment.elem_type, NextChar::Space);
   } else {
     WriteInitExpr(segment.offset);
   }
-  for (const Var& var : segment.vars) {
-    WriteVar(var, NextChar::Space);
+  for (const ElemExpr& expr : segment.elem_exprs) {
+    if (segment.passive) {
+      if (expr.kind == ElemExprKind::RefNull) {
+        WriteOpenSpace("ref.null");
+        WriteCloseSpace();
+      } else {
+        WriteOpenSpace("ref.func");
+        WriteVar(expr.var, NextChar::Space);
+        WriteCloseSpace();
+      }
+    } else {
+      assert(expr.kind == ElemExprKind::RefFunc);
+      WriteVar(expr.var, NextChar::Space);
+    }
   }
   WriteCloseNewline();
   elem_segment_index_++;
@@ -1552,9 +1561,8 @@ void WatWriter::WriteImport(const Import& import) {
       WriteCloseSpace();
       break;
 
-    case ExternalKind::Except:
-      WriteBeginException(cast<ExceptionImport>(&import)->except);
-      WriteCloseSpace();
+    case ExternalKind::Event:
+      WriteEvent(cast<EventImport>(&import)->event);
       break;
   }
 
@@ -1613,8 +1621,8 @@ Result WatWriter::WriteModule(const Module& module) {
       case ModuleFieldType::Import:
         WriteImport(*cast<ImportModuleField>(&field)->import);
         break;
-      case ModuleFieldType::Except:
-        WriteException(cast<ExceptionModuleField>(&field)->except);
+      case ModuleFieldType::Event:
+        WriteEvent(cast<EventModuleField>(&field)->event);
         break;
       case ModuleFieldType::Export:
         WriteExport(cast<ExportModuleField>(&field)->export_);
@@ -1684,8 +1692,8 @@ void WatWriter::BuildInlineExportMap() {
         index = module_->GetGlobalIndex(export_->var);
         break;
 
-      case ExternalKind::Except:
-        index = module_->GetExceptIndex(export_->var);
+      case ExternalKind::Event:
+        index = module_->GetEventIndex(export_->var);
         break;
     }
 
@@ -1729,8 +1737,8 @@ bool WatWriter::IsInlineExport(const Export& export_) {
       index = module_->GetGlobalIndex(export_.var);
       break;
 
-    case ExternalKind::Except:
-      index = module_->GetExceptIndex(export_.var);
+    case ExternalKind::Event:
+      index = module_->GetEventIndex(export_.var);
       break;
   }
 
