@@ -62,6 +62,8 @@ NAMED_VALUES = {
     'ELEM': 9,
     'CODE': 10,
     'DATA': 11,
+    'DATACOUNT': 12,
+    'EVENT': 13,
 
     # name subsection codes
     'NAME_MODULE': 0,
