@@ -124,6 +124,7 @@ def main(args):
   parser.add_argument('--enable-multi-value', action='store_true')
   parser.add_argument('--enable-bulk-memory', action='store_true')
   parser.add_argument('--enable-tail-call', action='store_true')
+  parser.add_argument('--enable-reference-types', action='store_true')
   parser.add_argument('--inline-exports', action='store_true')
   parser.add_argument('--inline-imports', action='store_true')
   parser.add_argument('file', help='test file.')
@@ -143,6 +144,7 @@ def main(args):
       '--enable-threads': options.enable_threads,
       '--enable-bulk-memory': options.enable_bulk_memory,
       '--enable-tail-call': options.enable_tail_call,
+      '--enable-reference-types': options.enable_reference_types,
       '--no-check': options.no_check,
   })
 
@@ -159,6 +161,7 @@ def main(args):
       '--enable-simd': options.enable_simd,
       '--enable-bulk-memory': options.enable_bulk_memory,
       '--enable-tail-call': options.enable_tail_call,
+      '--enable-reference-types': options.enable_reference_types,
       '--enable-threads': options.enable_threads,
       '--inline-exports': options.inline_exports,
       '--inline-imports': options.inline_imports,
